"""
Material Class
----------------


Description
==================================================

This class contains the data class to create a material library. internal use only

========================================================

"""
from collections import defaultdict, OrderedDict
from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from ..application.DataHandlers import dict2arg, arg2dict

class MatProperties(object):
    """list of constants of all possible materials constant names and how they are mapped to XML
    internalame=named used in script
    xmlname=named used in XML syntax

    Parameters
    ----------

    Returns
    -------

    """
    aedtname =     ['permittivity', 'permeability', 'conductivity', 'dielectric_loss_tangent', 'magnetic_loss_tangent', 'thermal_conductivity', 'mass_density', 'specific_heat', 'thermal_expansion_coefficient', 'youngs_modulus',   'poissons_ratio',   'diffusivity', 'molecular_mass', 'viscosity', 'core_loss_kh', 'core_loss_kc', 'core_loss_ke']
    defaultvalue = [1.0,             1.0,            0,              0,                         0,                       0.01,                      0,              0,               0,                               0,                  0,                  0.8,         0,                   0,                 0,                   0,                      0,                          0]
    defaultunit  = [None,            None,          '[siemens m^-1]', None,                     None,                   '[W m^-1 C^-1]',        '[Kg m^-3]',    '[J Kg^-1 C^-1]', '[C^-1]',                       '[Pa]',             None,               None,               None,               None,               None,           None,                  None,                       None]
    diel_order =   [3, 0, 1, 4, 5, 6, 7, 8, 9, 10, 11, 1]
    cond_order =   [2, 0, 1, 4, 5, 6, 7, 8, 9, 10, 11, 3]

    @classmethod
    def get_defaultunit(cls, aedtname=None):
        """Get the defaultunit for a given fullname or catname

        Parameters
        ----------
        aedtname : str
            aedtname (Default value = None)

        Returns
        -------
        str
            defaultunit if exists

        """
        if aedtname:
            return cls.defaultunit[cls.aedtname.index(aedtname)]
        else:
            raise TypeError("get_defaultunit: either fullname or catname MUST be defined")


    @classmethod
    def get_defaultvalue(cls, aedtname):
        """Get the defaultunit for a given fullname or catname

        Parameters
        ----------
        aedtname : str
             aedtname (Default value = None)

        Returns
        -------
        float
            defaultunit

        """
        if aedtname:
            return cls.defaultvalue[cls.aedtname.index(aedtname)]
        else:
            raise TypeError("get_defaultunit: either fullname or catname MUST be defined")

class SurfMatProperties(object):
    """list of constants of all possible materials constant names and how they are mapped to XML
    internalame=named used in script
    xmlname=named used in XML syntax

    Parameters
    ----------

    Returns
    -------

    """
    aedtname =     ['surface_emissivity', 'surface_roughness', 'surface_diffuse_absorptance', 'surface_incident_absorptance']
    defaultvalue = [1.0,             0,            0.4,              0.4]
    defaultunit  = [None,            '[m]',          None, None]

    @classmethod
    def get_defaultunit(cls, aedtname=None):
        """Get the defaultunit for a given fullname or catname

        Parameters
        ----------
        aedtname : str
            optional aedtname (Default value = None)

        Returns
        -------
        str
            defaultunit if exists

        """
        if aedtname:
            return cls.defaultunit[cls.aedtname.index(aedtname)]
        else:
            raise TypeError("get_defaultunit: either fullname or catname MUST be defined")


    @classmethod
    def get_defaultvalue(cls, aedtname=None):
        """Get the defaultunit for a given fullname or catname

        Parameters
        ----------
        aedtname : str
            optional aedtname (Default value = None)

        Returns
        -------
        float
            defaultunit

        """
        if aedtname:
            return cls.defaultvalue[cls.aedtname.index(aedtname)]
        else:
            raise TypeError("get_defaultunit: either fullname or catname MUST be defined")


class ClosedFormTM(object):
    """Class to manage Closed Form Terhmal Modifier."""
    Tref = '22cel'
    C1 = 0
    C2 = 0
    TL = '-273.15cel'
    TU = '1000cel'
    autocalculation = True
    TML = 1000
    TMU = 1000

class Dataset(object):
    """Data Class for DataSet Management"""
    ds = []
    unitx = ""
    unity = ""
    unitz = ""
    type = "Absolute"
    namex = ""
    namey = ""
    namez = None

class BasicValue(object):
    value = None
    dataset = None
    thermalmodifier = None



class MatProperty(object):
    """type: simple, anisotropic, tensor"""

    @property
    def _messenger(self):
        return self._parent._messenger

    def __init__(self, parent, name, val=None, thermalmodifier=None):
        self._parent = parent
        self._type = "simple"
        self.name =name
        self._property_value = [BasicValue()]
        self._unit = None
        if val is not None and isinstance(val,(str, float, int)):
            self.value = val
        elif val is not None and val["property_type"] == "AnisoProperty":
            self.type = "anisotropic"
            self.value = [val["component1"],val["component2"], val["component3"]]
        if not isinstance(thermalmodifier, list):
            thermalmodifier = [thermalmodifier]
        for tm in thermalmodifier:
            if tm:
                if tm["use_free_form"]:
                    self._property_value[tm["Index:"]].thermalmodifier = tm["free_form_value"]
                else:
                    self._property_value[tm["Index:"]].thermalmodifier = ClosedFormTM()
                    self._property_value[tm["Index:"]].thermalmodifier.Tref = tm["Tref"]
                    self._property_value[tm["Index:"]].thermalmodifier.C1 = tm["C1"]
                    self._property_value[tm["Index:"]].thermalmodifier.C2 = tm["C2"]
                    self._property_value[tm["Index:"]].thermalmodifier.TL = tm["TL"]
                    self._property_value[tm["Index:"]].thermalmodifier.TU = tm["TU"]
                    self._property_value[tm["Index:"]].thermalmodifier.autocalculation = tm["auto_calculation"]

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        """Set Material Property Type

        Parameters
        ----------
        type :
            simple", "nonlinear", "anisotropic", "tensor"

        Returns
        -------

        """
        self._type = type
        if self._type == "simple":
            self._property_value = [self._property_value[0]]
        elif self._type == "anisotropic":
            self._property_value = [self._property_value[0] for i in range(3)]
        elif self._type == "tensor":
            self._property_value = [self._property_value[0] for i in range(9)]
        elif self._type == "nonlinear":
            self._property_value = [self._property_value[0]]

    @property
    def value(self):
        if len(self._property_value) == 1:
            return self._property_value[0].value
        else:
            return [i.value for i in self._property_value]

    @value.setter
    def value(self, val):
        if isinstance(val, list):
            i = 0
            for el in val:
                if i >= len(self._property_value):
                    self._property_value.append(BasicValue())

                self._property_value[i].value = el
                i += 1
        else:
            self._property_value[0].value = val

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit):
        self._unit = unit

    @property
    def data_set(self):
        if len(self._property_value) == 1:
            return self._property_value[0].dataset
        else:
            return [i.dataset for i in self._property_value]

    @property
    def thermalmodifier(self):
        if len(self._property_value) == 1:
            return self._property_value[0].thermalmodifier
        else:
            return [i.thermalmodifier for i in self._property_value]

    def _add_thermal_modifier(self,formula, index):
        if "ModifierData" not in self._parent._props:
            tm = OrderedDict({'Property:': self.name, 'Index:': index, "prop_modifier": "thermal_modifier",
                              "use_free_form": True, "free_form_value": formula})
            self._parent._props["ModifierData"] = OrderedDict({"ThermalModifierData": OrderedDict(
                {"modifier_data": 'thermal_modifier_data',
                 "all_thermal_modifiers": OrderedDict({"one_thermal_modifier": tm})})})
        else:
            for tmname in self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"]:
                if isinstance(self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname], list):
                    found = False
                    for tm in self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname]:
                        if self.name == tm['Property:'] and index == tm['Index:']:
                            found = True
                            tm['use_free_form'] = True
                            tm['free_form_value'] = formula
                            tm.pop("Tref", None)
                            tm.pop("C1", None)
                            tm.pop("C2", None)
                            tm.pop("TL", None)
                            tm.pop("TU", None)
                    if not found:
                        tm = OrderedDict({'Property:': self.name, 'Index:': index, "prop_modifier": "thermal_modifier",
                                          "use_free_form": True, "free_form_value": formula})
                        self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].append(tm)
                elif self.name ==self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname][
                            'Property:'] and index ==self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname][
                            'Index:']:
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname]['use_free_form'] = True
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname]['free_form_value'] = formula
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].pop("Tref", None)
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].pop("C1", None)
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].pop("C2", None)
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].pop("TL", None)
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].pop("TU", None)

                else:
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname] = [
                        self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname]]
                    tm = OrderedDict({'Property:': self.name, 'Index:': index, "prop_modifier": "thermal_modifier",
                                      "use_free_form": True, "free_form_value": formula})
                    self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname].append(tm)
        return self._parent.update()

    def add_thermal_modifier_free_form(self, formula, index=0):
        """
        Add a Thermal modifier to a material Property using an free Form formula

        Parameters
        ----------
        formula : str
            Full Formula to apply
        index : int, Default ``0``

        Returns
        -------
        bool

        Examples
        --------

        >>> from pyaedt import Hfss
        >>> hfss = Hfss(specified_version="2021.1")
        >>> mat1 = hfss.materials.add_material("new_copper2")
        >>> mat1.aadd_thermal_modifier_free_form("if(Temp > 1000cel, 1, if(Temp < -273.15cel, 1, 1))")
        """
        self._property_value[index].thermalmodifier = formula
        return self._add_thermal_modifier(formula, index)


    def add_thermal_modifier_dataset(self, dataset_name, index=0):
        """
        Add a Thermal modifier to a material Property using an existing Dataset

        Parameters
        ----------
        dataset_name : str
            Project Dataset name
        index : int, Default ``0``

        Returns
        -------
        bool

        Examples
        --------

        >>> from pyaedt import Hfss
        >>> hfss = Hfss(specified_version="2021.1")
        >>> mat1 = hfss.materials.add_material("new_copper2")
        >>> mat1.add_thermal_modifier_dataset("$ds1")
        """

        formula = "pwl({}, Temp)".format(dataset_name)
        self._property_value[index].thermalmodifier = formula
        self._add_thermal_modifier(formula,index)

    def add_thermal_modifier_closed_form(self, tref=22, c1=0.0001, c2=1e-6, tl=-273.15, tu=1000, units="cel",
                                         auto_calc=True, tml=1000, tmu=1000, index=0):
        """
        Add a Thermal modifier to a material Property using Closed Form

        Parameters
        ----------
        tref : float
            Reference Temperature
        c1 : float
        c2 : float
        tl : float
        tu : float
        units : str, Default ``cel``
        auto_calc : bool, Default ``True``
        tml : float
        tmu : float
        index : int, Default ``0``

        Returns
        -------
        bool

        Examples
        --------

        >>> from pyaedt import Hfss
        >>> hfss = Hfss(specified_version="2021.1")
        >>> mat1 = hfss.materials.add_material("new_copper2")
        >>> mat1.permittivity.add_thermal_modifier_closed_form(c1 = 1e-3)
        """

        if index> len(self._property_value):
            self._messenger.add_error_message(
                "Wrong index number. Index must be 0 for simple or nonlinear properties, <=2 for anisotropic materials, <=9 for Tensors")
            return False
        self._property_value[index].thermalmodifier = ClosedFormTM()
        self._property_value[index].thermalmodifier.Tref = str(tref)+units
        self._property_value[index].thermalmodifier.C1 = str(c1)
        self._property_value[index].thermalmodifier.C2 = str(c2)
        self._property_value[index].thermalmodifier.TL = str(tl)+units
        self._property_value[index].thermalmodifier.TU = str(tu)+units
        self._property_value[index].thermalmodifier.autocalculation = auto_calc
        if not auto_calc:
            self._property_value[index].thermalmodifier.TML = tml
            self._property_value[index].thermalmodifier.TMU = tmu
        if auto_calc:
            tm_new = OrderedDict({'Property:': self.name, 'Index:': index, "prop_modifier": "thermal_modifier",
                              "use_free_form": False, "Tref": str(tref) + units,
                              "C1": str(c1), "C2": str(c2), "TL": str(tl) + units, "TU": str(tu) + units,
                              "auto_calculation": True})
        else:
            tm_new = OrderedDict({'Property:': self.name, 'Index:': index, "prop_modifier": "thermal_modifier",
                              "use_free_form": False, "Tref": str(tref) + units,
                              "C1": str(c1), "C2": str(c2), "TL": str(tl) + units, "TU": str(tu) + units,
                              "auto_calculation": False, "TML":str(tml), "TMU":str(tmu)})
        if "ModifierData" not in self._parent._props:
            self._parent._props["ModifierData"] = OrderedDict({"ThermalModifierData": OrderedDict(
                {"modifier_data": 'thermal_modifier_data',
                 "all_thermal_modifiers": OrderedDict({"one_thermal_modifier": tm_new})})})
        else:
            for tmname in self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"]:
                tml = self._parent._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][tmname]
                if isinstance(tml,  list):
                    found = False
                    for tm in tml:
                        if self.name == tm['Property:'] and index == tm['Index:']:
                            found = True
                            tm['use_free_form'] = False
                            tm.pop("free_form_value", None)
                            tm['Tref'] = str(tref)+units
                            tm['C1'] = str(c1)
                            tm['C2'] = str(c2)
                            tm['TL'] = str(tl)+units
                            tm['TU'] = str(tu)+units
                            tm["auto_calculation"] = auto_calc
                            if auto_calc:
                                tm["TML"] = tml
                                tm["TMU"] = tmu
                            else:
                                tm.pop("TML", None)
                                tm.pop("TMU", None)
                    if not found:
                        tml.append(tm_new)
                elif self.name == tml['Property:'] and index == tml['Index:']:
                    tml['use_free_form'] = False
                    tml.pop("free_form_value", None)
                    tml['Tref'] = str(tref)+units
                    tml['C1'] = str(c1)
                    tml['C2'] = str(c1)
                    tml['TL'] = str(tl)+units
                    tml['TU'] = str(tl)+units
                    tml['auto_calculation'] = auto_calc
                    if not auto_calc:
                        tml["TML"] = str(tml)
                        tml["TMU"] = str(tmu)
                    else:
                        tml.pop("TML", None)
                        tml.pop("TMU", None)
                else:
                    tml = [tml]
                    tml.append(tm_new)
        return self._parent.update()

class CommonMaterial(object):
    """Class for Frequency Dependence Datasets"""

    @property
    def odefinition_manager(self):
        """:return: Definition Manager"""
        return self._parent.oproject.GetDefinitionManager()

    @property
    def _omaterial_manager(self):
        """:return:Material Manager"""
        return self.odefinition_manager.GetManager("Material")

    @property
    def _messenger(self):
        """ """
        return self._parent._messenger

    @property
    def oproject(self):
        """ """
        return self._parent._oproject

    @property
    def desktop(self):
        """ """
        return self._parent._desktop

    def __init__(self, parent, name, props=None):
        self._parent = parent
        self.name = name
        self.coordinate_system = ""
        if props:
            self._props = props
        else:
            self._props = OrderedDict()
        if "CoordinateSystemType" in self._props:
            self.coordinate_system = self._props["CoordinateSystemType"]
        else:
            self._props["CoordinateSystemType"] = "Cartesian"
            self.coordinate_system = "Cartesian"
        if "BulkOrSurfaceType" in self._props:
            self.bulkorsurface = self._props["BulkOrSurfaceType"]
        else:
            self._props["BulkOrSurfaceType"] = 1
        if "ModTime" in self._props:
            self._modtime = self._props["ModTime"]
            del self._props["ModTime"]
        if "LibLocation" in self._props:
            self.lib_location = self._props["LibLocation"]
            del self._props["LibLocation"]
        if "ModSinceLib" in self._props:
            self.mod_since_lib = self._props["ModSinceLib"]
            del self._props["ModSinceLib"]


    @aedt_exception_handler
    def _get_args(self, props=None):
        if not props:
            props = self._props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg

    def _update_props(self, propname, provpavlue, update_aedt=True):
        if isinstance(provpavlue, list) and self.__dict__["_"+propname].type != "simple" and self.__dict__["_"+propname].type != "nonlinear":
                i=1
                for val in provpavlue:
                    self._props[propname]["component"+str(i)] = str(val)
                    i += 1
                if update_aedt:
                    return self.update()
        elif  isinstance(provpavlue, (str, float, int)):
            self._props[propname] = str(provpavlue)
            if update_aedt:
                return self.update()
        else:
            return False


class Material(CommonMaterial, object):
    """Class for Material Properties"""

    def __init__(self, parent, name, props=None):
        CommonMaterial.__init__(self, parent, name, props)
        self.thermal_material_type = "Solid"
        if "thermal_material_type" in self._props:
            self.thermal_material_type = self._props["thermal_material_type"]["Choice"]
        if "PhysicsTypes" in self._props:
            self.physics_type = self._props["PhysicsTypes"]["set"]
        else:
            self.physics_type = ['Electromagnetic', 'Thermal', 'Structural']
            self._props["PhysicsTypes"] = OrderedDict({"set":['Electromagnetic', 'Thermal', 'Structural']})

        for property in MatProperties.aedtname:
            if property in self._props:
                mods = None
                if "ModifierData" in self._props:
                    for mod in self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"]:
                        if isinstance(self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][mod],
                                      list):
                            for one_tm in self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][
                                mod]:
                                if one_tm["Property:"] == property:
                                    if mods:
                                        mods = [mods]
                                        mods.append(one_tm)
                                    else:
                                        mods = one_tm
                        else:
                            if self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][mod][
                                "Property:"] == property:
                                mods = self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][mod]
                self.__dict__["_" + property] = MatProperty(self, property, self._props[property], mods)
            else:
                self.__dict__["_" + property] = MatProperty(self, property,
                                                            MatProperties.get_defaultvalue(aedtname=property), None)
        pass

    @property
    def permittivity(self):
        """

        Returns
        -------
        MatProperty
            Permittivity Value of the material
        """
        return self._permittivity

    @permittivity.setter
    def permittivity(self, value):

        self._permittivity.value = value
        self._update_props("permittivity", value)

    @property
    def permeability(self):
        """

        Returns
        -------
        MatProperty
            Permeability Value of the material
        """
        return self._permeability

    @permeability.setter
    def permeability(self, value):

        self._permeability.value = value
        self._update_props("permeability", value)

    @property
    def conductivity(self):
        """

        Returns
        -------
        MatProperty
            Conductivity Value of the material
        """
        return self._conductivity

    @conductivity.setter
    def conductivity(self, value):
        self._conductivity.value = value
        self._update_props("conductivity", value)

    @property
    def dielectric_loss_tangent(self):
        """

        Returns
        -------
        MatProperty
            Dielectric Loss Tangent Value of the material
        """
        return self._dielectric_loss_tangent

    @dielectric_loss_tangent.setter
    def dielectric_loss_tangent(self, value):

        self._dielectric_loss_tangent.value = value
        self._update_props("dielectric_loss_tangent", value)

    @property
    def magnetic_loss_tangent(self):
        """

        Returns
        -------
        MatProperty
            Magnetic Loss Tangent Value of the material
        """
        return self._magnetic_loss_tangent

    @magnetic_loss_tangent.setter
    def magnetic_loss_tangent(self, value):

        self._magnetic_loss_tangent.value = value
        self._update_props("magnetic_loss_tangent", value)
    @property
    def thermal_conductivity(self):
        """

        Returns
        -------
        MatProperty
            Thermal Conductivity Value of the material

        """
        return self._thermal_conductivity

    @thermal_conductivity.setter
    def thermal_conductivity(self, value):

        self._thermal_conductivity.value = value
        self.physics_type = ['Electromagnetic', 'Thermal', 'Structural']
        self._props["PhysicsTypes"] = OrderedDict({"set": ['Electromagnetic', 'Thermal', 'Structural']})
        self._update_props("thermal_conductivity", value)

    @property
    def mass_density(self):
        """

        Returns
        -------
        MatProperty
            Mass Density Value of the material

        """
        return self._mass_density

    @mass_density.setter
    def mass_density(self, value):

        self._mass_density.value = value
        self._update_props("mass_density", value)

    @property
    def specific_heat(self):
        """

        Returns
        -------
        MatProperty
            Specific Heat Value of the material

        """
        return self._specific_heat

    @specific_heat.setter
    def specific_heat(self, value):

        self._specific_heat.value = value
        self._update_props("specific_heat", value)

    @property
    def thermal_expansion_coefficient(self):
        """

        Returns
        -------
        MatProperty
            Thermal Expansion Coefficient  of the material

        """
        return self._thermal_expansion_coefficient

    @thermal_expansion_coefficient.setter
    def thermal_expansion_coefficient(self, value):

        self._thermal_expansion_coefficient.value = value
        self._update_props("thermal_expansion_coefficient", value)

    @property
    def youngs_modulus(self):
        """

        Returns
        -------
        MatProperty
            Youngs Modulus  of the material

        """
        return self._youngs_modulus

    @youngs_modulus.setter
    def youngs_modulus(self, value):
        self._youngs_modulus.value = value
        self.physics_type = ['Electromagnetic', 'Thermal', 'Structural']
        self._props["PhysicsTypes"] = OrderedDict({"set": ['Electromagnetic', 'Thermal', 'Structural']})
        self._update_props("youngs_modulus", value)

    @property
    def poissons_ratio(self):
        """

        Returns
        -------
        MatProperty
            Poisson Ratio's of the material
        """
        return self._poissons_ratio

    @poissons_ratio.setter
    def poissons_ratio(self, value):
        self._poissons_ratio.value = value
        self.physics_type = ['Electromagnetic', 'Thermal', 'Structural']
        self._props["PhysicsTypes"] = OrderedDict({"set": ['Electromagnetic', 'Thermal', 'Structural']})
        self._update_props("poissons_ratio", value)

    @property
    def diffusivity(self):
        return self._diffusivity

    @diffusivity.setter
    def diffusivity(self, value):
        self._diffusivity.value = value
        self._update_props("diffusivity", value)

    @property
    def molecular_mass(self):
        return self._molecular_mass

    @molecular_mass.setter
    def molecular_mass(self, value):
        self._molecular_mass.value = value
        self._update_props("molecular_mass", value)

    @property
    def viscosity(self):
        return self._viscosity

    @viscosity.setter
    def viscosity(self, value):
        self._viscosity.value = value
        self._update_props("viscosity", value)

    @property
    def core_loss_kh(self):
        return self._core_loss_kh

    @core_loss_kh.setter
    def core_loss_kh(self, value):
        self._core_loss_kh.value = value
        self._update_props("core_loss_kh", value)

    @property
    def core_loss_kc(self):
        return self._core_loss_kc

    @core_loss_kc.setter
    def core_loss_kc(self, value):
        self._core_loss_kc.value = value
        self._update_props("core_loss_kc", value)

    @property
    def core_loss_ke(self):
        return self._core_loss_ke

    @core_loss_ke.setter
    def core_loss_ke(self, value):
        self._core_loss_ke.value = value
        self._update_props("core_loss_ke", value)

    def is_conductor(self, threshold=100000):
        """Check if material is conductor

        Parameters
        ----------
        threshold : float
            Threshold to define if material is condcutor or not. Default is ``100000``

        Returns
        -------
        bool
            ``True`` if material is conductor.   Material is defined as conductor if cond>=100000. Threshold can be modified
        """
        cond = self.conductivity.value
        if not cond:
            return False
        if "Freq" in str(cond):
            return True
        try:
            if float(cond) >= threshold:
                return True
        except:
            return False
        return False

    @aedt_exception_handler
    def is_dielectric(self, threshold=100000):
        """Check if material is dielectric

        Parameters
        ----------
        threshold : float
            Threshold to define if material is dielectric. Default conductivitivy value is ``100000``

        Returns
        -------
        bool
            ``True`` if material is dielectric.   Material is defined as conductor if conductivity>=100000.
            Threshold can be modified
        """
        return not self.is_conductor()

    @aedt_exception_handler
    def update(self):
        """Update material in AEDT


        Returns
        -------
        bool
            ``True`` if succeeded

        """

        args = self._get_args()
        if self._does_material_exists(self.name):
            self.odefinition_manager.EditMaterial(self.name, args)
        else:
            self.odefinition_manager.AddMaterial(args)
        return True

    @aedt_exception_handler
    def _does_material_exists(self, material_name):
        listmatprj = [i.lower() for i in list(self.odefinition_manager.GetProjectMaterialNames())]
        if material_name.lower() in listmatprj:
            return True
        else:
            return False



class SurfaceMaterial(CommonMaterial, object):

    def __init__(self, parent, name, props=None):
        CommonMaterial.__init__(self,parent, name, props)
        self.surface_clarity_type = "Opaque"
        if "surface_clarity_type" in self._props:
            self.surface_clarity_type = self._props["surface_clarity_type"]["Choice"]
        if "PhysicsTypes" in self._props:
            self.physics_type = self._props["PhysicsTypes"]["set"]
        else:
            self.physics_type = ['Thermal']
        for property in SurfMatProperties.aedtname:
            if property in self._props:
                mods = None
                if "ModifierData" in self._props:
                    for mod in self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"]:
                        if isinstance(self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][mod],
                                      list):
                            for one_tm in self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][
                                mod]:
                                if one_tm["Property:"] == property:
                                    if mods:
                                        mods = [mods]
                                        mods.append(one_tm)
                                    else:
                                        mods = one_tm
                        else:
                            if self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][mod][
                                "Property:"] == property:
                                mods = self._props["ModifierData"]["ThermalModifierData"]["all_thermal_modifiers"][mod]
                self.__dict__["_" + property] = MatProperty(self, property, self._props[property], mods)
            else:
                self.__dict__["_" + property] = MatProperty(self, property, SurfMatProperties.get_defaultvalue(aedtname=property))
        pass

    @property
    def emissivity(self):
        return self._surface_emissivity

    @emissivity.setter
    def emissivity(self, value):

        self._surface_emissivity.value = value
        self._update_props("surface_emissivity", value)

    @property
    def surface_diffuse_absorptance(self):
        return self._surface_diffuse_absorptance

    @surface_diffuse_absorptance.setter
    def surface_diffuse_absorptance(self, value):

        self._surface_diffuse_absorptance.value = value
        self._update_props("surface_diffuse_absorptance", value)

    @property
    def surface_incident_absorptance(self):
        return self._surface_incident_absorptance

    @surface_incident_absorptance.setter
    def surface_incident_absorptance(self, value):

        self._surface_incident_absorptance.value = value
        self._update_props("surface_incident_absorptance", value)

    @property
    def surface_roughness(self):
        return self._surface_roughness

    @surface_roughness.setter
    def surface_roughness(self, value):

        self._surface_roughness.value = value
        self._update_props("surface_roughness", value)

    @aedt_exception_handler
    def update(self):
        """Update material in AEDT


        Returns
        -------
        type
            Bool

        """
        args = self._get_args()
        if self._does_material_exists(self.name):
            self.odefinition_manager.EditSurfaceMaterial(self.name, args)
        else:
            self.odefinition_manager.AddSurfaceMaterial(args)
        return True

    @aedt_exception_handler
    def _does_material_exists(self, szMat):
        a = self.odefinition_manager.DoesSurfaceMaterialExist(szMat)
        if a!=0:
            return True
        return False



from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from ..application.DataHandlers import dict2arg
from collections import OrderedDict

meshlink = [("ImportMesh", False)]
autosweep = [("RangeType", "LinearStep"), ("RangeStart", "1GHz"), ("RangeEnd", "10GHz"), ("RangeStep", "1GHz")]
autosweeps = [("Sweep", autosweep)]
multifreq = [("1GHz", [0.02]), ("2GHz", [0.02]), ("5GHz", [0.02])]
sweepsbr = [("RangeType", "LinearStep"), ("RangeStart", "1GHz"),
            ("RangeEnd", "10GHz"), ("RangeStep", "1GHz")]
sweepssbr = [("Sweep", sweepsbr)]
muoption = [("MuNonLinearBH", True)]
transientelectrostatic = [("SaveField", True), ("Stop", "100s"), ("InitialStep", "0.01s"),
                          ("MaxStep", "5s")]
transienthfss = [("TimeProfile", "Broadband Pulse"), ("HfssFrequency", "1GHz"),
                 ("MinFreq", "100MHz"), ("MaxFreq", "1GHz"),
                 ("NumFreqsExtracted", 401), ("SweepMinFreq", "100MHz"),
                 ("SweepMaxFreq", "1GHz"), ("UseAutoTermination", 1),
                 ("SteadyStateCriteria", 0.01), ("UseMinimumDuration", 0),
                 ("TerminateOnMaximum", 0)]
HFSSDrivenAuto = [("IsEnabled", True), ("MeshLink", meshlink), ("AutoSolverSetting", "Balanced"),
                  ("Sweeps", autosweeps),
                  ("SaveRadFieldsOnly", False), ("SaveAnyFields", True), ("Type", "Discrete")]
HFSSDrivenDefault = [("AdaptMultipleFreqs", False), ("MultipleAdaptiveFreqsSetup", multifreq), ("Frequency", "5GHz"),
                     ("MaxDeltaS", 0.02), ("PortsOnly", False),
                     ("UseMatrixConv", False), ("MaximumPasses", 6), ("MinimumPasses", 1),
                     ("MinimumConvergedPasses", 1),
                     ("PercentRefinement", 30), ("IsEnabled", True), ("MeshLink", meshlink), ("BasisOrder", 1),
                     ("DoLambdaRefine", True), ("DoMaterialLambda", True), ("SetLambdaTarget", False),
                     ("Target", 0.3333),
                     ("UseMaxTetIncrease", False), ("PortAccuracy", 2), ("UseABCOnPort", False), ("SetPortMinMaxTri",
                                                                                                  False),
                     ("UseDomains", False), ("UseIterativeSolver", False), ("SaveRadFieldsOnly", False),
                     ("SaveAnyFields", True),
                     ("IESolverType", "Auto"), ("LambdaTargetForIESolver", 0.15),
                     ("UseDefaultLambdaTgtForIESolver", True),
                     ("IE Solver Accuracy", "Balanced")]
HFSSEigen = [("MinimumFrequency", "2GHz"), ("NumModes", 1), ("MaxDeltaFreq", 10), ("ConvergeOnRealFreq", False),
             ("MaximumPasses", 3), ("MinimumPasses", 1), ("MinimumConvergedPasses", 1),
             ("PercentRefinement", 30), ("IsEnabled", True), ("MeshLink", meshlink),
             ("BasisOrder", 1), ("DoLambdaRefine", True), ("DoMaterialLambda", True),
             ("SetLambdaTarget", False), ("Target", 0.2), ("UseMaxTetIncrease", False)]
HFSSTransient = [("Frequency", "5GHz"), ("MaxDeltaS", 0.02), ("MaximumPasses", 20), ("UseImplicitSolver", True),
                 ("IsEnabled", True), ("MeshLink", meshlink), ("BasisOrder", -1), ("Transient", transienthfss)]
HFSSSBR = [("IsEnabled", True), ("MeshLink", meshlink), ("IsSbrRangeDoppler", False),
           ("RayDensityPerWavelength", 4), ("MaxNumberOfBounces", 5), ("RadiationSetup", ""),
           ("PTDUTDSimulationSettings", "None"), ("Sweeps", sweepssbr), ("ComputeFarFields", False)]
MaxwellTransient = [("Enabled", True), ("MeshLink", meshlink), ("NonlinearSolverResidual", "0.005"),
                    ("ScalarPotential", "Second Order"), ("SmoothBHCurve", False), ("StopTime", "10000000ns"),
                    ("TimeStep", "2000000ns"),
                    ("OutputError", False), ("UseControlProgram", False), ("ControlProgramName", ""),
                    ("ControlProgramArg", ""), ("CallCtrlProgAfterLastStep", False),
                    ("FastReachSteadyState", False), ("AutoDetectSteadyState", False),
                    ("IsGeneralTransient", True), ("IsHalfPeriodicTransient", False),
                    ("SaveFieldsType", "None"), ("CacheSaveKind", "Count"), ("NumberSolveSteps", 1),
                    ("RangeStart", "0s"), ("RangeEnd", "0.1s")]
Magnetostatic = [("Enabled", True), ("MeshLink", meshlink), ("MaximumPasses", 10), ("MinimumPasses", 2),
                 ("MinimumConvergedPasses", 1), ("PercentRefinement", 30), ("SolveFieldOnly", False),
                 ("PercentError", 1), ("SolveMatrixAtLast", True), ("PercentError", 1),
                 ("UseIterativeSolver", False), ("RelativeResidual", 1E-06), ("NonLinearResidual", 0.001),
                 ("SmoothBHCurve", False), ("MuOption", muoption)]
Electrostatic = [("Enabled", True), ("MeshLink", meshlink), ("MaximumPasses", 10), ("MinimumPasses", 2),
                 ("MinimumConvergedPasses", 1), ("PercentRefinement", 30), ("SolveFieldOnly", False),
                 ("PercentError", 1), ("SolveMatrixAtLast", True), ("PercentError", 1),
                 ("UseIterativeSolver", False), ("RelativeResidual", 1E-06), ("NonLinearResidual", 0.001)]
EddyCurrent = [("Enabled", True), ("MeshLink", meshlink), ("MaximumPasses", 6), ("MinimumPasses", 1),
               ("MinimumConvergedPasses", 1), ("PercentRefinement", 30), ("SolveFieldOnly", False), ("PercentError", 1),
               ("SolveMatrixAtLast", True), ("PercentError", 1), ("UseIterativeSolver", False),
               ("RelativeResidual", 1E-5), ("NonLinearResidual", 0.0001), ("SmoothBHCurve", False),
               ("Frequency", "60Hz"), ("HasSweepSetup", False), ("SweepSetupType", "LinearStep"),
               ("StartValue", "1e-08GHz"), ("StopValue", "1e-06GHz"), ("StepSize", "1e-08GHz"),
               ("UseHighOrderShapeFunc", False), ("UseMuLink", False)]
ElectricTransient = [("Enabled",), ("MeshLink", meshlink), ("Tolerance", 0.005), ("ComputePowerLoss", False),
                     ("Data", transientelectrostatic),
                     ("Initial Voltage", "0mV")]
SteadyTemperatureAndFlow = [("Enabled", True), ("Flow Regime", "Laminar"), ("Include Temperature", True),
                            ("Include Flow", True), ("Include Gravity", False),
                            ("Solution Initialization - X Velocity", "0m_per_sec"),
                            ("Solution Initialization - Y Velocity", "0m_per_sec"),
                            ("Solution Initialization - Z Velocity", "0m_per_sec"),
                            ("Solution Initialization - Temperature", "AmbientTemp"),
                            ("Solution Initialization - Turbulent Kinetic Energy", "1m2_per_s2"),
                            ("Solution Initialization - Turbulent Dissipation Rate", "1m2_per_s3"),
                            ("Solution Initialization - Specific Dissipation Rate", "1diss_per_s"),
                            ("Convergence Criteria - Flow", "0.001"),
                            ("Convergence Criteria - Energy", "1e-07"),
                            ("Convergence Criteria - Turbulent Kinetic Energy", "0.001"),
                            ("Convergence Criteria - Turbulent Dissipation Rate", "0.001"),
                            ("Convergence Criteria - Specific Dissipation Rate", "0.001"),
                            ("Convergence Criteria - Discrete Ordinates", "1e-06"),
                            ("IsEnabled", False), ("Radiation Model", "Off"),
                            ("Under-relaxation - Pressure", "0.7"), ("Under-relaxation - Momentum", "0.3"),
                            ("Under-relaxation - Temperature", "1"),
                            ("Under-relaxation - Turbulent Kinetic Energy", "0.8"),
                            ("Under-relaxation - Turbulent Dissipation Rate", "0.8"),
                            ("Under-relaxation - Specific Dissipation Rate", "0.8"),
                            ("Discretization Scheme - Pressure", "Standard"),
                            ("Discretization Scheme - Momentum", "First"),
                            ("Discretization Scheme - Temperature", "Second"), ("Secondary Gradient", False),
                            ("Discretization Scheme - Turbulent Kinetic Energy", "First"),
                            ("Discretization Scheme - Turbulent Dissipation Rate", "First"),
                            ("Discretization Scheme - Specific Dissipation Rate", "First"),
                            ("Discretization Scheme - Discrete Ordinates", "First"),
                            ("Linear Solver Type - Pressure", "V"),
                            ("Linear Solver Type - Momentum", "flex"),
                            ("Linear Solver Type - Temperature", "F"),
                            ("Linear Solver Type - Turbulent Kinetic Energy", "flex"),
                            ("Linear Solver Type - Turbulent Dissipation Rate", "flex"),
                            ("Linear Solver Type - Specific Dissipation Rate", "flex"),
                            ("Linear Solver Termination Criterion - Pressure", "0.1"),
                            ("Linear Solver Termination Criterion - Momentum", "0.1"),
                            ("Linear Solver Termination Criterion - Temperature", "0.1"),
                            ("Linear Solver Termination Criterion - Turbulent Kinetic Energy", "0.1"),
                            ("Linear Solver Termination Criterion - Turbulent Dissipation  Rate", "0.1"),
                            ("Linear Solver Termination Criterion - Specific Dissipation Rate", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Pressure", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Momentum", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Temperature", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Turbulent Kinetic Energy", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Turbulent Dissipation Rate", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Specific Dissipation Rate", "0.1"),
                            ("Linear Solver Stabilization - Pressure", "None"),
                            ("Linear Solver Stabilization - Temperature", "None"),
                            ("Frozen Flow Simulation", False),
                            ("Sequential Solve of Flow and Energy Equations", False),
                            ("Convergence Criteria - Max Iterations", 100)]
SteadyTemperatureOnly = [("Enabled", True), ("Flow Regime", "Laminar"), ("Include Temperature", True),
                         ("Include Gravity", False),
                         ("Solution Initialization - X Velocity", "0m_per_sec"),
                         ("Solution Initialization - Y Velocity", "0m_per_sec"),
                         ("Solution Initialization - Z Velocity", "0m_per_sec"),
                         ("Solution Initialization - Temperature", "AmbientTemp"),
                         ("Solution Initialization - Turbulent Kinetic Energy", "1m2_per_s2"),
                         ("Solution Initialization - Turbulent Dissipation Rate", "1m2_per_s3"),
                         ("Solution Initialization - Specific Dissipation Rate", "1diss_per_s"),
                         ("Convergence Criteria - Flow", "0.001"),
                         ("Convergence Criteria - Energy", "1e-07"),
                         ("Convergence Criteria - Turbulent Kinetic Energy", "0.001"),
                         ("Convergence Criteria - Turbulent Dissipation Rate", "0.001"),
                         ("Convergence Criteria - Specific Dissipation Rate", "0.001"),
                         ("Convergence Criteria - Discrete Ordinates", "1e-06"),
                         ("IsEnabled", False), ("Radiation Model", "Off"),
                         ("Under-relaxation - Pressure", "0.7"), ("Under-relaxation - Momentum", "0.3"),
                         ("Under-relaxation - Temperature", "1"),
                         ("Under-relaxation - Turbulent Kinetic Energy", "0.8"),
                         ("Under-relaxation - Turbulent Dissipation Rate", "0.8"),
                         ("Under-relaxation - Specific Dissipation Rate", "0.8"),
                         ("Discretization Scheme - Pressure", "Standard"),
                         ("Discretization Scheme - Momentum", "First"),
                         ("Discretization Scheme - Temperature", "Second"), ("Secondary Gradient", False),
                         ("Discretization Scheme - Turbulent Kinetic Energy", "First"),
                         ("Discretization Scheme - Turbulent Dissipation Rate", "First"),
                         ("Discretization Scheme - Specific Dissipation Rate", "First"),
                         ("Discretization Scheme - Discrete Ordinates", "First"),
                         ("Linear Solver Type - Pressure", "V"),
                         ("Linear Solver Type - Momentum", "flex"),
                         ("Linear Solver Type - Temperature", "F"),
                         ("Linear Solver Type - Turbulent Kinetic Energy", "flex"),
                         ("Linear Solver Type - Turbulent Dissipation Rate", "flex"),
                         ("Linear Solver Type - Specific Dissipation Rate", "flex"),
                         ("Linear Solver Termination Criterion - Pressure", "0.1"),
                         ("Linear Solver Termination Criterion - Momentum", "0.1"),
                         ("Linear Solver Termination Criterion - Temperature", "0.1"),
                         ("Linear Solver Termination Criterion - Turbulent Kinetic Energy", "0.1"),
                         ("Linear Solver Termination Criterion - Turbulent Dissipation  Rate", "0.1"),
                         ("Linear Solver Termination Criterion - Specific Dissipation Rate", "0.1"),
                         ("Linear Solver Residual Reduction Tolerance - Pressure", "0.1"),
                         ("Linear Solver Residual Reduction Tolerance - Momentum", "0.1"),
                         ("Linear Solver Residual Reduction Tolerance - Temperature", "0.1"),
                         ("Linear Solver Residual Reduction Tolerance - Turbulent Kinetic Energy", "0.1"),
                         ("Linear Solver Residual Reduction Tolerance - Turbulent Dissipation Rate", "0.1"),
                         ("Linear Solver Residual Reduction Tolerance - Specific Dissipation Rate", "0.1"),
                         ("Linear Solver Stabilization - Pressure", "None"),
                         ("Linear Solver Stabilization - Temperature", "None"),
                         ("Sequential Solve of Flow and Energy Equations", False),
                         ("Convergence Criteria - Max Iterations", 100)]
SteadyFlowOnly = [("Enabled", True), ("Flow Regime", "Laminar"), ("Include Flow", True), ("Include Gravity", False),
                  ("Solution Initialization - X Velocity", "0m_per_sec"),
                  ("Solution Initialization - Y Velocity", "0m_per_sec"),
                  ("Solution Initialization - Z Velocity", "0m_per_sec"),
                  ("Solution Initialization - Temperature", "AmbientTemp"),
                  ("Solution Initialization - Turbulent Kinetic Energy", "1m2_per_s2"),
                  ("Solution Initialization - Turbulent Dissipation Rate", "1m2_per_s3"),
                  ("Solution Initialization - Specific Dissipation Rate", "1diss_per_s"),
                  ("Convergence Criteria - Flow", "0.001"),
                  ("Convergence Criteria - Energy", "1e-07"),
                  ("Convergence Criteria - Turbulent Kinetic Energy", "0.001"),
                  ("Convergence Criteria - Turbulent Dissipation Rate", "0.001"),
                  ("Convergence Criteria - Specific Dissipation Rate", "0.001"),
                  ("Convergence Criteria - Discrete Ordinates", "1e-06"),
                  ("IsEnabled", False), ("Radiation Model", "Off"),
                  ("Under-relaxation - Pressure", "0.7"), ("Under-relaxation - Momentum", "0.3"),
                  ("Under-relaxation - Temperature", "1"), ("Under-relaxation - Turbulent Kinetic Energy", "0.8"),
                  ("Under-relaxation - Turbulent Dissipation Rate", "0.8"),
                  ("Under-relaxation - Specific Dissipation Rate", "0.8"),
                  ("Discretization Scheme - Pressure", "Standard"), ("Discretization Scheme - Momentum", "First"),
                  ("Discretization Scheme - Temperature", "First"), ("Secondary Gradient", False),
                  ("Discretization Scheme - Turbulent Kinetic Energy", "First"),
                  ("Discretization Scheme - Turbulent Dissipation Rate", "First"),
                  ("Discretization Scheme - Specific Dissipation Rate", "First"),
                  ("Discretization Scheme - Discrete Ordinates", "First"),
                  ("Linear Solver Type - Pressure", "V"),
                  ("Linear Solver Type - Momentum", "flex"),
                  ("Linear Solver Type - Temperature", "F"),
                  ("Linear Solver Type - Turbulent Kinetic Energy", "flex"),
                  ("Linear Solver Type - Turbulent Dissipation Rate", "flex"),
                  ("Linear Solver Type - Specific Dissipation Rate", "flex"),
                  ("Linear Solver Termination Criterion - Pressure", "0.1"),
                  ("Linear Solver Termination Criterion - Momentum", "0.1"),
                  ("Linear Solver Termination Criterion - Temperature", "0.1"),
                  ("Linear Solver Termination Criterion - Turbulent Kinetic Energy", "0.1"),
                  ("Linear Solver Termination Criterion - Turbulent Dissipation  Rate", "0.1"),
                  ("Linear Solver Termination Criterion - Specific Dissipation Rate", "0.1"),
                  ("Linear Solver Residual Reduction Tolerance - Pressure", "0.1"),
                  ("Linear Solver Residual Reduction Tolerance - Momentum", "0.1"),
                  ("Linear Solver Residual Reduction Tolerance - Temperature", "0.1"),
                  ("Linear Solver Residual Reduction Tolerance - Turbulent Kinetic Energy", "0.1"),
                  ("Linear Solver Residual Reduction Tolerance - Turbulent Dissipation Rate", "0.1"),
                  ("Linear Solver Residual Reduction Tolerance - Specific Dissipation Rate", "0.1"),
                  ("Linear Solver Stabilization - Pressure", "None"),
                  ("Linear Solver Stabilization - Temperature", "None"),
                  ("Frozen Flow Simulation", False),
                  ("Sequential Solve of Flow and Energy Equations", False),
                  ("Convergence Criteria - Max Iterations", 100)]
Q3DCond = [("MaxPass", 10), ("MinPass", 1), ("MinConvPass", 1), ("PerError", 1), ("PerRefine", 30)]
Q3DMult = [("MaxPass", 1), ("MinPass", 1), ("MinConvPass", 1), ("PerError", 1), ("PerRefine", 30)]
Q3DDC = [("SolveResOnly", False), ("Cond", Q3DCond), ("Mult", Q3DMult)]
Q3DCap = [("MaxPass", 10), ("MinPass", 1), ("MinConvPass", 1), ("PerError", 1),
          ("PerRefine", 30), ("AutoIncreaseSolutionOrder", True), ("SolutionOrder", "High"),
          ("Solver Type", "Iterative")]
Q3DAC = [("MaxPass", 10), ("MinPass", 1), ("MinConvPass", 1), ("PerError", 1), ("PerRefine", 30)]
Matrix = [("AdaptiveFreq", "1GHz"), ("SaveFields", False), ("Enabled", True), ("Cap", Q3DCap), ("DC", Q3DDC),
          ("AC", Q3DAC)]
OutputQuantities = []
NoiseOutputQuantities = []
SweepDefinition = [("Variable", "Freq"), ("Data", "LINC 1GHz 5GHz 501"), ("OffsetF1", False),
                   ("Synchronize", 0)]
NexximLNA = [("DataBlockID", 16), ("OptionName", "(Default Options)"), ("AdditionalOptions", ""),
             ("AlterBlockName", ""), ("FilterText", ""), ("AnalysisEnabled", 1), ("OutputQuantities", OutputQuantities),
             ("NoiseOutputQuantities", NoiseOutputQuantities), ("Name", "LinearFrequency"),
             ("LinearFrequencyData", [False, 0.1, False, "", False]),
             ("SweepDefinition", SweepDefinition)]
NexximDC = [("DataBlockID", 15), ("OptionName", "(Default Options)"), ("AdditionalOptions", ""),
            ("AlterBlockName", ""), ("FilterText", ""), ("AnalysisEnabled", 1), ("OutputQuantities", OutputQuantities),
            ("NoiseOutputQuantities", NoiseOutputQuantities), ("Name", "LinearFrequency")]
NexximTransient = [("DataBlockID", 10), ("OptionName", "(Default Options)"), ("AdditionalOptions", ""),
                   ("AlterBlockName", ""), ("FilterText", ""), ("AnalysisEnabled", 1),
                   ("OutputQuantities", OutputQuantities), ("NoiseOutputQuantities", NoiseOutputQuantities),
                   ("Name", "LinearFrequency"),
                   ("TransientData", ["0.1ns", "10ns"]), ("TransientNoiseData", [False, "", "", 0, 1, 0, False, 1]),
                   ("TransientOtherData", ["default"])]
NexximQuickEye = []
NexximVerifEye = []
NexximAMI = []
NexximOscillatorRSF = []
NexximOscillator1T = []
NexximOscillatorNT = []
NexximHarmonicBalance1T = []
NexximHarmonicBalanceNT = []
NexximSystem = []
NexximTVNoise = []
HSPICE = []

HFSS3DLayout_Properties = [("Enable", "true")]
HFSS3DLayout_AdvancedSettings = [
    ("AccuracyLevel", 2),
    ("GapPortCalibration", True),
    ("ReferenceLengthRatio", 0.25),
    ("RefineAreaRatio", 4),
    ("DRCOn", False),
    ("FastSolverOn", False),
    ("StartFastSolverAt", 3000),
    ("LoopTreeOn", True),
    ("SingularElementsOn", False),
    ("UseStaticPortSolver", False),
    ("UseThinMetalPortSolver", False),
    ("ComputeBothEvenAndOddCPWModes", False),
    ("ZeroMetalLayerThickness", 0),
    ("ThinDielectric", 0),
    ("UseShellElements", False),
    ("SVDHighCompression", False),
    ("NumProcessors", 1),
    ("UseHfssIterativeSolver", False),
    ("UseHfssMUMPSSolver", True),
    ("RelativeResidual", 1E-06),
    ("EnhancedLowFreqAccuracy", False),
    ("OrderBasis", -1),
    ("MaxDeltaZo", 2),
    ("UseRadBoundaryOnPorts", False),
    ("SetTrianglesForWavePort", False),
    ("MinTrianglesForWavePort", 100),
    ("MaxTrianglesForWavePort", 500),
    ("numprocessorsdistrib", 1),
    ("CausalMaterials", True),
    ("enabledsoforopti", True),
    ("usehfsssolvelicense", False),
    ("ExportAfterSolve", False),
    ("ExportDir", ""),
    ("CircuitSparamDefinition", False),
    ("CircuitIntegrationType", "FFT"),
    ("DesignType", "generic"),
    ("MeshingMethod", "Phi"),
    ("EnableDesignIntersectionCheck", True)]
HFSS3DLayout_CurveApproximation = [
    ("ArcAngle", "30deg"),
    ("StartAzimuth", "0deg"),
    ("UseError", False),
    ("Error", "0meter"),
    ("MaxPoints", 8),
    ("UnionPolys", True),
    ("Replace3DTriangles", True)]
HFSS3DLayout_Q3D_DCSettings = [
    ("SolveResOnly", True),
    ("Cond", Q3DCond),
    ("Mult", Q3DMult),
    ("Solution Order", "Normal")]
# HFSS3DLayout_AdaptiveFrequencyData = [
#     ("AdaptiveFrequency", "5GHz"),
#     ("MaxDelta", "0.02"),
#     ("MaxPasses", 10),
#     ("Expressions", [])]
CGDataBlock = [("MaxPass", 10),
               ("MinPass", 1),
               ("MinConvPass", 1),
               ("PerError", 1),
               ("PerRefine", 30),
               ("DataType", "CG"),
               ("Included", True),
               ("UseParamConv", True),
               ("UseLossyParamConv", False),
               ("PerErrorParamConv", 1),
               ("UseLossConv", True)]
RLDataBlock = [("MaxPass", 10),
               ("MinPass", 1),
               ("MinConvPass", 1),
               ("PerError", 1),
               ("PerRefine", 30),
               ("DataType", "CG"),
               ("Included", True),
               ("UseParamConv", True),
               ("UseLossyParamConv", False),
               ("PerErrorParamConv", 1),
               ("UseLossConv", True)]
Open = [("AdaptiveFreq", "1GHz"), ("SaveFields", True), ("Enabled", True), ("MeshLink", meshlink),
        ("CGDataBlock", CGDataBlock), ("RLDataBlock", RLDataBlock), ("CacheSaveKind", "Delta"), ("ConstantDelta", "0s")]

Close = [("AdaptiveFreq", "1GHz"), ("SaveFields", True), ("Enabled", True), ("MeshLink", meshlink),
         ("CGDataBlock", CGDataBlock), ("RLDataBlock", RLDataBlock), ("CacheSaveKind", "Delta"),
         ("ConstantDelta", "0s")]

TransientTemperatureAndFlow = [("Enabled", True), ("Flow Regime", "Laminar"), ("Include Temperature", True),
                               ("Include Flow", True), ("Include Gravity", False), ("Include Solar", False),
                               ("Solution Initialization - X Velocity", "0m_per_sec"),
                               ("Solution Initialization - Y Velocity", "0m_per_sec"),
                               ("Solution Initialization - Z Velocity", "0m_per_sec"),
                               ("Solution Initialization - Temperature", "AmbientTemp"),
                               ("Solution Initialization - Turbulent Kinetic Energy", "1m2_per_s2"),
                               ("Solution Initialization - Turbulent Dissipation Rate", "1m2_per_s3"),
                               ("Solution Initialization - Specific Dissipation Rate", "1diss_per_s"),
                               ("Solution Initialization - Use Model Based Flow Initialization", False),
                               ("Convergence Criteria - Flow", "0.001"),
                               ("Convergence Criteria - Energy", "1e-07"),
                               ("Convergence Criteria - Turbulent Kinetic Energy", "0.001"),
                               ("Convergence Criteria - Turbulent Dissipation Rate", "0.001"),
                               ("Convergence Criteria - Specific Dissipation Rate", "0.001"),
                               ("Convergence Criteria - Discrete Ordinates", "1e-06"), ("IsEnabled:=", False),
                               ("Radiation Model", "Off"),
                               ("Solar Radiation Model", "Solar Radiation Calculator"),
                               ("Solar Radiation - Scattering Fraction", "0"),
                               ("Solar Radiation - Day", 1), ("Solar Radiation - Month", 1),
                               ("Solar Radiation - Hours", 0),
                               ("Solar Radiation - Minutes", 0), ("Solar Radiation - GMT", "0"),
                               ("Solar Radiation - Latitude", "0"),
                               ("Solar Radiation - Latitude Direction", "East"), ("Solar Radiation - Longitude", "0"),
                               ("Solar Radiation - Longitude Direction", "North"),
                               ("Solar Radiation - Ground Reflectance", "0"),
                               ("Solar Radiation - Sunshine Fraction", "0"), ("Solar Radiation - North X", "0"),
                               ("Solar Radiation - North Y", "0"), ("Solar Radiation - North Z", "1"),
                               ("Under-relaxation - Pressure", "0.3"),
                               ("Under-relaxation - Momentum", "0.7"), ("Under-relaxation - Temperature", "1"),
                               ("Under-relaxation - Turbulent Kinetic Energy", "0.8"),
                               ("Under-relaxation - Turbulent Dissipation Rate", "0.8"),
                               ("Under-relaxation - Specific Dissipation Rate", "0.8"),
                               ("Discretization Scheme - Pressure", "Standard"),
                               ("Discretization Scheme - Momentum", "First"),
                               ("Discretization Scheme - Temperature", "First"), ("Secondary Gradient", False),
                               ("Discretization Scheme - Turbulent Kinetic Energy", "First"),
                               ("Discretization Scheme - Turbulent Dissipation Rate", "First"),
                               ("Discretization Scheme - Specific Dissipation Rate", "First"),
                               ("Discretization Scheme - Discrete Ordinates", "First"),
                               ("Linear Solver Type - Pressure", "V"), ("Linear Solver Type - Momentum", "flex"),
                               ("Linear Solver Type - Temperature", "F"),
                               ("Linear Solver Type - Turbulent Kinetic Energy", "flex"),
                               ("Linear Solver Type - Turbulent Dissipation Rate", "flex"),
                               ("Linear Solver Type - Specific Dissipation Rate", "flex"),
                               ("Linear Solver Termination Criterion - Pressure", "0.1"),
                               ("Linear Solver Termination Criterion - Momentum", "0.1"),
                               ("Linear Solver Termination Criterion - Temperature", "0.1"),
                               ("Linear Solver Termination Criterion - Turbulent Kinetic Energy", "0.1"),
                               ("Linear Solver Termination Criterion - Turbulent Dissipation Rate", "0.1"),
                               ("Linear Solver Termination Criterion - Specific Dissipation Rate", "0.1"),
                               ("Linear Solver Residual Reduction Tolerance - Pressure", "0.1"),
                               ("Linear Solver Residual Reduction Tolerance - Momentum", "0.1"),
                               ("Linear Solver Residual Reduction Tolerance - Temperature", "0.1"),
                               ("Linear Solver Residual Reduction Tolerance - Turbulent Kinetic Energy", "0.1"),
                               ("Linear Solver Residual Reduction Tolerance - Turbulent Dissipation Rate", "0.1"),
                               ("Linear Solver Residual Reduction Tolerance - Specific Dissipation Rate", "0.1"),
                               ("Linear Solver Stabilization - Pressure", "None"),
                               ("Linear Solver Stabilization - Temperature", "None"),
                               ("Coupled pressure-velocity formulation", False),
                               ("Frozen Flow Simulation", False), ("Start Time:=", "0s"), ("Stop Time:=", "20s"),
                               ("Time Step:=", "1s"), ("Iterations per Time Step", 20), ("Import Start Time", False),
                               ("Copy Fields From Source", False), ("SaveFieldsType", "Every N Steps"),
                               ("N Steps:=", "10s"), ("Enable Control Program", False), ("Control Program Name", "")]

TransientTemperatureOnly = [("Enabled", True), ("Flow Regime", "Laminar"), ("Include Temperature", True),
                            ("Include Flow", False), ("Include Gravity", False), ("Include Solar", False),
                            ("Solution Initialization - X Velocity", "0m_per_sec"),
                            ("Solution Initialization - Y Velocity", "0m_per_sec"),
                            ("Solution Initialization - Z Velocity", "0m_per_sec"),
                            ("Solution Initialization - Temperature", "AmbientTemp"),
                            ("Solution Initialization - Turbulent Kinetic Energy", "1m2_per_s2"),
                            ("Solution Initialization - Turbulent Dissipation Rate", "1m2_per_s3"),
                            ("Solution Initialization - Specific Dissipation Rate", "1diss_per_s"),
                            ("Solution Initialization - Use Model Based Flow Initialization", False),
                            ("Convergence Criteria - Flow", "0.001"),
                            ("Convergence Criteria - Energy", "1e-07"),
                            ("Convergence Criteria - Turbulent Kinetic Energy", "0.001"),
                            ("Convergence Criteria - Turbulent Dissipation Rate", "0.001"),
                            ("Convergence Criteria - Specific Dissipation Rate", "0.001"),
                            ("Convergence Criteria - Discrete Ordinates", "1e-06"), ("IsEnabled:=", False),
                            ("Radiation Model", "Off"),
                            ("Solar Radiation Model", "Solar Radiation Calculator"),
                            ("Solar Radiation - Scattering Fraction", "0"),
                            ("Solar Radiation - Day", 1), ("Solar Radiation - Month", 1),
                            ("Solar Radiation - Hours", 0),
                            ("Solar Radiation - Minutes", 0), ("Solar Radiation - GMT", "0"),
                            ("Solar Radiation - Latitude", "0"),
                            ("Solar Radiation - Latitude Direction", "East"), ("Solar Radiation - Longitude", "0"),
                            ("Solar Radiation - Longitude Direction", "North"),
                            ("Solar Radiation - Ground Reflectance", "0"),
                            ("Solar Radiation - Sunshine Fraction", "0"), ("Solar Radiation - North X", "0"),
                            ("Solar Radiation - North Y", "0"), ("Solar Radiation - North Z", "1"),
                            ("Under-relaxation - Pressure", "0.3"),
                            ("Under-relaxation - Momentum", "0.7"), ("Under-relaxation - Temperature", "1"),
                            ("Under-relaxation - Turbulent Kinetic Energy", "0.8"),
                            ("Under-relaxation - Turbulent Dissipation Rate", "0.8"),
                            ("Under-relaxation - Specific Dissipation Rate", "0.8"),
                            ("Discretization Scheme - Pressure", "Standard"),
                            ("Discretization Scheme - Momentum", "First"),
                            ("Discretization Scheme - Temperature", "First"), ("Secondary Gradient", False),
                            ("Discretization Scheme - Turbulent Kinetic Energy", "First"),
                            ("Discretization Scheme - Turbulent Dissipation Rate", "First"),
                            ("Discretization Scheme - Specific Dissipation Rate", "First"),
                            ("Discretization Scheme - Discrete Ordinates", "First"),
                            ("Linear Solver Type - Pressure", "V"), ("Linear Solver Type - Momentum", "flex"),
                            ("Linear Solver Type - Temperature", "F"),
                            ("Linear Solver Type - Turbulent Kinetic Energy", "flex"),
                            ("Linear Solver Type - Turbulent Dissipation Rate", "flex"),
                            ("Linear Solver Type - Specific Dissipation Rate", "flex"),
                            ("Linear Solver Termination Criterion - Pressure", "0.1"),
                            ("Linear Solver Termination Criterion - Momentum", "0.1"),
                            ("Linear Solver Termination Criterion - Temperature", "0.1"),
                            ("Linear Solver Termination Criterion - Turbulent Kinetic Energy", "0.1"),
                            ("Linear Solver Termination Criterion - Turbulent Dissipation Rate", "0.1"),
                            ("Linear Solver Termination Criterion - Specific Dissipation Rate", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Pressure", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Momentum", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Temperature", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Turbulent Kinetic Energy", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Turbulent Dissipation Rate", "0.1"),
                            ("Linear Solver Residual Reduction Tolerance - Specific Dissipation Rate", "0.1"),
                            ("Linear Solver Stabilization - Pressure", "None"),
                            ("Linear Solver Stabilization - Temperature", "None"),
                            ("Coupled pressure-velocity formulation", False),
                            ("Frozen Flow Simulation", False), ("Start Time:=", "0s"), ("Stop Time:=", "20s"),
                            ("Time Step:=", "1s"), ("Iterations per Time Step", 20), ("Import Start Time", False),
                            ("Copy Fields From Source", False), ("SaveFieldsType", "Every N Steps"),
                            ("N Steps:=", "10s"), ("Enable Control Program", False), ("Control Program Name", "")]

TransientFlowOnly = [("Enabled", True), ("Flow Regime", "Laminar"), ("Include Temperature", False),
                     ("Include Flow", True), ("Include Gravity", False), ("Include Solar", False),
                     ("Solution Initialization - X Velocity", "0m_per_sec"),
                     ("Solution Initialization - Y Velocity", "0m_per_sec"),
                     ("Solution Initialization - Z Velocity", "0m_per_sec"),
                     ("Solution Initialization - Temperature", "AmbientTemp"),
                     ("Solution Initialization - Turbulent Kinetic Energy", "1m2_per_s2"),
                     ("Solution Initialization - Turbulent Dissipation Rate", "1m2_per_s3"),
                     ("Solution Initialization - Specific Dissipation Rate", "1diss_per_s"),
                     ("Solution Initialization - Use Model Based Flow Initialization", False),
                     ("Convergence Criteria - Flow", "0.001"),
                     ("Convergence Criteria - Energy", "1e-07"),
                     ("Convergence Criteria - Turbulent Kinetic Energy", "0.001"),
                     ("Convergence Criteria - Turbulent Dissipation Rate", "0.001"),
                     ("Convergence Criteria - Specific Dissipation Rate", "0.001"),
                     ("Convergence Criteria - Discrete Ordinates", "1e-06"), ("IsEnabled:=", False),
                     ("Radiation Model", "Off"),
                     ("Solar Radiation Model", "Solar Radiation Calculator"),
                     ("Solar Radiation - Scattering Fraction", "0"),
                     ("Solar Radiation - Day", 1), ("Solar Radiation - Month", 1),
                     ("Solar Radiation - Hours", 0),
                     ("Solar Radiation - Minutes", 0), ("Solar Radiation - GMT", "0"),
                     ("Solar Radiation - Latitude", "0"),
                     ("Solar Radiation - Latitude Direction", "East"), ("Solar Radiation - Longitude", "0"),
                     ("Solar Radiation - Longitude Direction", "North"),
                     ("Solar Radiation - Ground Reflectance", "0"),
                     ("Solar Radiation - Sunshine Fraction", "0"), ("Solar Radiation - North X", "0"),
                     ("Solar Radiation - North Y", "0"), ("Solar Radiation - North Z", "1"),
                     ("Under-relaxation - Pressure", "0.3"),
                     ("Under-relaxation - Momentum", "0.7"), ("Under-relaxation - Temperature", "1"),
                     ("Under-relaxation - Turbulent Kinetic Energy", "0.8"),
                     ("Under-relaxation - Turbulent Dissipation Rate", "0.8"),
                     ("Under-relaxation - Specific Dissipation Rate", "0.8"),
                     ("Discretization Scheme - Pressure", "Standard"),
                     ("Discretization Scheme - Momentum", "First"),
                     ("Discretization Scheme - Temperature", "First"), ("Secondary Gradient", False),
                     ("Discretization Scheme - Turbulent Kinetic Energy", "First"),
                     ("Discretization Scheme - Turbulent Dissipation Rate", "First"),
                     ("Discretization Scheme - Specific Dissipation Rate", "First"),
                     ("Discretization Scheme - Discrete Ordinates", "First"),
                     ("Linear Solver Type - Pressure", "V"), ("Linear Solver Type - Momentum", "flex"),
                     ("Linear Solver Type - Temperature", "F"),
                     ("Linear Solver Type - Turbulent Kinetic Energy", "flex"),
                     ("Linear Solver Type - Turbulent Dissipation Rate", "flex"),
                     ("Linear Solver Type - Specific Dissipation Rate", "flex"),
                     ("Linear Solver Termination Criterion - Pressure", "0.1"),
                     ("Linear Solver Termination Criterion - Momentum", "0.1"),
                     ("Linear Solver Termination Criterion - Temperature", "0.1"),
                     ("Linear Solver Termination Criterion - Turbulent Kinetic Energy", "0.1"),
                     ("Linear Solver Termination Criterion - Turbulent Dissipation Rate", "0.1"),
                     ("Linear Solver Termination Criterion - Specific Dissipation Rate", "0.1"),
                     ("Linear Solver Residual Reduction Tolerance - Pressure", "0.1"),
                     ("Linear Solver Residual Reduction Tolerance - Momentum", "0.1"),
                     ("Linear Solver Residual Reduction Tolerance - Temperature", "0.1"),
                     ("Linear Solver Residual Reduction Tolerance - Turbulent Kinetic Energy", "0.1"),
                     ("Linear Solver Residual Reduction Tolerance - Turbulent Dissipation Rate", "0.1"),
                     ("Linear Solver Residual Reduction Tolerance - Specific Dissipation Rate", "0.1"),
                     ("Linear Solver Stabilization - Pressure", "None"),
                     ("Linear Solver Stabilization - Temperature", "None"),
                     ("Coupled pressure-velocity formulation", False),
                     ("Frozen Flow Simulation", False), ("Start Time:=", "0s"), ("Stop Time:=", "20s"),
                     ("Time Step:=", "1s"), ("Iterations per Time Step", 20), ("Import Start Time", False),
                     ("Copy Fields From Source", False), ("SaveFieldsType", "Every N Steps"),
                     ("N Steps:=", "10s"), ("Enable Control Program", False), ("Control Program Name", "")]


def HFSS3DLayout_AdaptiveFrequencyData(freq):
    """

    Parameters
    ----------
    freq :
        

    Returns
    -------

    """
    value = [
        ("AdaptiveFrequency", freq),
        ("MaxDelta", "0.02"),
        ("MaxPasses", 10),
        ("Expressions", [], None)]
    return value


HFSS3DLayout_SingleFrequencyDataList = [
    ("AdaptiveFrequencyData", HFSS3DLayout_AdaptiveFrequencyData("5GHz"))]
HFSS3DLayout_BroadbandFrequencyDataList = [
    ("AdaptiveFrequencyData", HFSS3DLayout_AdaptiveFrequencyData("5GHz")),
    ("AdaptiveFrequencyData", HFSS3DLayout_AdaptiveFrequencyData("10GHz"))]
HFSS3DLayout_MultiFrequencyDataList = [
    ("AdaptiveFrequencyData", HFSS3DLayout_AdaptiveFrequencyData("2.5GHz")),
    ("AdaptiveFrequencyData", HFSS3DLayout_AdaptiveFrequencyData("5GHz")),
    ("AdaptiveFrequencyData", HFSS3DLayout_AdaptiveFrequencyData("10GHz"))]
HFSS3DLayout_AdaptiveSettings = [
    ("DoAdaptive", True),
    ("SaveFields", False),
    ("SaveRadFieldsOnly", False),
    ("MaxRefinePerPass", 30),
    ("MinPasses", 1),
    ("MinConvergedPasses", 1),
    ("AdaptType", "kSingle"),  # possible values are "kSingle", "kMultiFrequencies", "kBroadband"
    ("Basic", True),
    ("SingleFrequencyDataList", HFSS3DLayout_SingleFrequencyDataList),
    ("BroadbandFrequencyDataList", HFSS3DLayout_BroadbandFrequencyDataList),
    ("MultiFrequencyDataList", HFSS3DLayout_MultiFrequencyDataList)]
HFSS3DLayout = [
    ("Properties", HFSS3DLayout_Properties),
    ("CustomSetup", False),
    ("SolveSetupType", "HFSS"),
    ("PercentRefinementPerPass", 30),
    ("MinNumberOfPasses", 1),
    ("MinNumberOfConvergedPasses", 1),
    ("UseDefaultLambda", True),
    ("UseMaxRefinement", False),
    ("MaxRefinement", 1000000),
    ("SaveAdaptiveCurrents", False),
    ("SaveLastAdaptiveRadFields", False),
    ("ProdMajVerID", -1),
    ("ProjDesignSetup", ""),
    ("ProdMinVerID", -1),
    ("Refine", False),
    ("Frequency", "10GHz"),
    ("LambdaRefine", True),
    ("MeshSizeFactor", 1.5),
    ("QualityRefine", True),
    ("MinAngle", "15deg"),
    ("UniformityRefine", False),
    ("MaxRatio", 2),
    ("Smooth", False),
    ("SmoothingPasses", 5),
    ("UseEdgeMesh", False),
    ("UseEdgeMeshAbsLength", False),
    ("EdgeMeshRatio", 0.1),
    ("EdgeMeshAbsLength", "1000mm"),
    ("LayerProjectThickness", "0meter"),
    ("UseDefeature", True),
    ("UseDefeatureAbsLength", False),
    ("DefeatureRatio", 1E-06),
    ("DefeatureAbsLength", "0mm"),
    ("InfArrayDimX", 0),
    ("InfArrayDimY", 0),
    ("InfArrayOrigX", 0),
    ("InfArrayOrigY", 0),
    ("InfArraySkew", 0),
    ("ViaNumSides", 6),
    ("ViaMaterial", "copper"),
    ("Style25DVia", "Mesh"),
    ("Replace3DTriangles", True),
    ("LayerSnapTol", "1e-05"),
    ("ViaDensity", 0),
    ("HfssMesh", True),
    ("Q3dPostProc", False),
    ("UnitFactor", 1000),
    ("Verbose", False),
    ("NumberOfProcessors", 0),
    ("SmallVoidArea", -2E-09),
    ("HealingOption", 1),
    ("InclBBoxOption", 1),
    ("AuxBlock", []),
    ("DoAdaptive", True),
    ("Color", ["R:=", 0, "G:=", 0, "B:=", 0], None),  # TODO: create something smart for color arrays, like a class
    ("AdvancedSettings", HFSS3DLayout_AdvancedSettings),
    ("CurveApproximation", HFSS3DLayout_CurveApproximation),
    ("Q3D_DCSettings", HFSS3DLayout_Q3D_DCSettings),
    ("AdaptiveSettings", HFSS3DLayout_AdaptiveSettings)]

MechTerm = [("Enabled", True), ("MeshLink", meshlink),
            ("Solver", "Program Controlled"), ("Stepping", "Program Controlled")]
MechModal = [("Enabled", True), ("MeshLink", meshlink),
             ("Max Modes", 6), ("Limit Search", True), ("Range Max", "100MHz"), ("Range Min", "0Hz"),
             ("Solver", "Program Controlled")]

MechStructural = [("Enabled", True), ("MeshLink", meshlink), ("Solver", "Program Controlled"),
                  ("Stepping", "Program Controlled")]

# TODO complete the list of templates for other Solvers

GRM = [("Enabled", True), ("MeshLink", meshlink),
       ("RatedOutputPower", "1W"), ("RatedVoltage", "208V"), ("RatedSpeed", "3600rpm"),
       ("OperatingTemperature", "75cel"),
       ("OperationType", "Motor"), ("LoadType", "ConstPower"), ("RatedPowerFactor", "0.8"), ("Frequency", "60Hz"),
       ("CapacitivePowerFactor", False)]

TR = []


class SweepHFSS(object):
    """ """

    def __init__(self, oanalysis, setupname, sweepname, sweeptype="Interpolating", props=None):
        self.oanalysis = oanalysis
        self.props = {}
        self.setupname = setupname
        self.name = sweepname
        if props:
            self.props = props
        else:
            self.setupname = setupname
            self.name = sweepname
            self.props["Type"] = sweeptype
            self.props["isenabled"] = True
            self.props["RangeType"] = "LinearCount"
            self.props["RangeStart"] = "2.5GHz"
            self.props["RangeEnd"] = "7.5GHz"
            self.props["SaveSingleField"] = False
            self.props["RangeCount"] = 401
            self.props["RangeStep"] = "1MHz"
            self.props["RangeSamples"] = 11
            self.props["SaveFields"] = True
            self.props["SaveRadFields"] = True
            self.props["GenerateFieldsForAllFreqs"] = False
            self.props["InterpTolerance"] = 0.5
            self.props["InterpMaxSolns"] = 250
            self.props["InterpMinSolns"] = 0
            self.props["InterpMinSubranges"] = 1
            self.props["InterpUseS"] = True
            self.props["InterpUsePortImped"] = False
            self.props["InterpUsePropConst"] = True
            self.props["UseDerivativeConvergence"] = False
            self.props["InterpDerivTolerance"] = 0.2
            self.props["EnforcePassivity"] = True
            self.props["UseFullBasis"] = True
            self.props["PassivityErrorTolerance"] = 0.0001
            self.props["EnforceCausality"] = False
            self.props["UseQ3DForDCSolve"] = False
            self.props["SMatrixOnlySolveMode"] = "Auto"
            self.props["SMatrixOnlySolveAbove"] = "1MHz"
            self.props["SweepRanges"] = []

    @aedt_exception_handler
    def add_subrange(self, rangetype, start, end, count):
        """Add Subrange to sweep

        Parameters
        ----------
        rangetype :
            subrange type
        start :
            freq start
        end :
            freq end
        count :
            freq count or freq step

        Returns
        -------
        type
            bool

        """
        range = {}
        range["RangeType"] = rangetype
        range["RangeStart"] = start
        range["RangeEnd"] = end
        if rangetype == "LinearCount":
            range["RangeCount"] = count
        elif rangetype == "LinearStep":
            range["RangeStep"] = count
        elif rangetype == "LogScale":
            range["RangeCount"] = self.props["RangeCount"]
            range["RangeSamples"] = count
        elif rangetype == "SinglePoints":
            range["SaveSingleField"] = False
        self.props["SweepRanges"].append(range)

    @aedt_exception_handler
    def create(self):
        """Create Sweep
        
        :return:Bool

        Parameters
        ----------

        Returns
        -------

        """
        self.oanalysis.InsertFrequencySweep(self.setupname, self._get_args())
        return True

    @aedt_exception_handler
    def update(self):
        """Update Sweep
        
        :return:Bool

        Parameters
        ----------

        Returns
        -------

        """
        self.oanalysis.EditFrequencySweep(self.setupname, self.name, self._get_args())
        return True

    @aedt_exception_handler
    def _get_args(self, props=None):
        """

        Parameters
        ----------
        props :
             (Default value = None)

        Returns
        -------

        """
        if props is None:
            props = self.props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg


class SweepHFSS3DLayout(object):
    """ """

    def __init__(self, oanalysis, setupname, sweepname, sweeptype="Interpolating", props=None):
        self.oanalysis = oanalysis
        self.props = {}
        self.setupname = setupname
        self.name = sweepname
        if props:
            self.props = props
        else:
            self.setupname = setupname
            self.name = sweepname

            self.props["Properties"] = OrderedDict({"Enable":True})
            self.props["Sweeps"] = OrderedDict(
                {"Variable": "Sweep 1", "Data": "LIN 1Hz 20GHz 0.05GHz", "OffsetF1": False, "Synchronize": 0})
            self.props["GenerateSurfaceCurrent"] = False
            self.props["SaveRadFieldsOnly"] = False
            if sweeptype == "Interpolating":
                self.props["FastSweep"] = True
            else:
                self.props["FastSweep"] = True

            self.props["SaveSingleField"] = False
            self.props["ZoSelected"] = False
            self.props["SAbsError"] = 0.005
            self.props["ZoPercentError"] = 1
            self.props["GenerateStateSpace"] = False
            self.props["EnforcePassivity"] = False
            self.props["PassivityTolerance"] = 0.0001
            self.props["UseQ3DForDC"] = False
            self.props["ResimulateDC"] = False
            self.props["MaxSolutions"] = 250
            self.props["InterpUseSMatrix"] = True
            self.props["InterpUsePortImpedance"] = True
            self.props["InterpUsePropConst"] = True
            self.props["InterpUseFullBasis"] = True
            self.props["AdvDCExtrapolation"] = False
            self.props["MinSolvedFreq"] = "0.01GHz"
            self.props["CustomFrequencyString"] = ""
            self.props["AllEntries"] = False
            self.props["AllDiagEntries"] = False
            self.props["AllOffDiagEntries"] = False
            self.props["MagMinThreshold"] = 0.01


    @aedt_exception_handler
    def add_subrange(self, rangetype, start, end, count):
        """Add Subrange to sweep

        Parameters
        ----------
        rangetype :
            subrange type
        start :
            freq start
        end :
            freq end
        count :
            freq count or freq step

        Returns
        -------
        type
            bool

        """
        range = ""
        if rangetype == "LinearCount":
            range = " LINC "+str(start)+" "+str(end) + " "+str(count)
        elif rangetype == "LinearStep":
            range = " LIN "+str(start)+" "+str(end) + " "+str(count)
        elif rangetype == "LogScale":
            range = " DEC "+str(start)+" "+str(end) + " "+str(count)
        self.props["Sweeps"]["Data"] += range
        return self.update()

    @aedt_exception_handler
    def change_range(self, rangetype, start, end, count):
        """Change Range of the sweep

        Parameters
        ----------
        rangetype :
            subrange type
        start :
            freq start
        end :
            freq end
        count :
            freq count or freq step

        Returns
        -------
        type
            bool

        """
        range = ""
        if rangetype == "LinearCount":
            range = "LINC "+str(start)+" "+str(end) + " "+str(count)
        elif rangetype == "LinearStep":
            range = "LIN "+str(start)+" "+str(end) + " "+str(count)
        elif rangetype == "LogScale":
            range = "DEC "+str(start)+" "+str(end) + " "+str(count)
        self.props["Sweeps"]["Data"] = range
        return self.update()

    @aedt_exception_handler
    def create(self):
        """Create Sweep

        :return:Bool

        Parameters
        ----------

        Returns
        -------

        """
        self.oanalysis.AddSweep(self.setupname, self._get_args())
        return True

    @aedt_exception_handler
    def update(self):
        """Update Sweep

        :return:Bool

        Parameters
        ----------

        Returns
        -------

        """
        self.oanalysis.EditSweep(self.setupname, self.name, self._get_args())
        return True

    @aedt_exception_handler
    def _get_args(self, props=None):
        """

        Parameters
        ----------
        props :
             (Default value = None)

        Returns
        -------

        """
        if props is None:
            props = self.props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg


class SweepQ3D(object):
    """ """

    def __init__(self, oanalysis, setupname, sweepname, sweeptype="Interpolating", props=None):
        self.oanalysis = oanalysis
        self.setupname = setupname
        self.name = sweepname
        self.props = {}
        if props:
            self.props = props
        else:
            self.props["Type"] = sweeptype
            self.props["isenabled"] = True
            self.props["RangeType"] = "LinearCount"
            self.props["RangeStart"] = "2.5GHz"
            self.props["RangeStep"] = "1GHz"
            self.props["RangeEnd"] = "7.5GHz"
            self.props["SaveSingleField"] = False

            self.props["RangeSamples"] = 3
            self.props["RangeCount"] = 401
            self.props["SaveFields"] = True
            self.props["SaveRadFields"] = True
            self.props["SweepRanges"] = []

    @aedt_exception_handler
    def add_subrange(self, type, start, end, count):
        """Add Subrange to sweep

        Parameters
        ----------
        type :
            subrange type
        start :
            freq start
        end :
            freq end
        count :
            freq count or freq step

        Returns
        -------
        type
            bool

        """
        range = {}
        range["RangeType"] = type
        range["RangeStart"] = start
        range["RangeEnd"] = end
        if type == "LinearCount":
            range["RangeCount"] = count
        elif type == "LinearStep":
            range["RangeStep"] = count
        elif type == "LogScale":
            range["RangeCount"] = self.props["RangeCount"]
            range["RangeSamples"] = count
        self.props["SweepRanges"].append(range)
        return True

    @aedt_exception_handler
    def create(self):
        """Create a Sweep
        
        :return: Bool

        Parameters
        ----------

        Returns
        -------

        """
        self.oanalysis.InsertSweep(self.setupname, self._get_args())
        return True

    @aedt_exception_handler
    def update(self):
        """Update Sweep
        
        :return: Bool

        Parameters
        ----------

        Returns
        -------

        """
        self.oanalysis.EditSweep(self.setupname, self.name, self._get_args())

        return True

    @aedt_exception_handler
    def _get_args(self, props=None):
        """

        Parameters
        ----------
        props :
             (Default value = None)

        Returns
        -------

        """
        if props is None:
            props = self.props
        arg = ["NAME:" + self.name]
        dict2arg(props, arg)
        return arg


class SetupKeys(object):
    """ """
    defaultSetups = {"DrivenModal": 1, "DrivenTerminal": 1, "Eigenmode": 2,
                     "Transient Network": 3, "SBR+": 4, "Transient": 5, "Magnetostatic": 6, "EddyCurrent": 7,
                     "Electrostatic": 8, "ElectroDCConduction": 9,
                     "ElectricTransient": 10, "SteadyTemperatureAndFlow": 11, "SteadyTemperatureOnly": 12,
                     "SteadyFlowOnly": 13,
                     "SteadyState": 11,
                     "Matrix": 14, "NexximLNA": 15, "NexximDC": 16, "NexximTransient": 17, "NexximQuickEye": 18,
                     "NexximVerifEye": 19, "NexximAMI": 20, "NexximOscillatorRSF": 21, "NexximOscillator1T": 22,
                     "NexximOscillatorNT": 23, "NexximHarmonicBalance1T": 24, "NexximHarmonicBalanceNT": 25,
                     "NexximSystem": 26, "NexximTVNoise": 27, "HSPICE": 28, "HFSS3DLayout": 29, "Open": 30, "Close": 31,
                     "Thermal": 32, "Modal": 33, "IRIM": 34, "ORIM": 34, "SRIM": 34, "WRIM": 34,
                     "DFIG": 34, "AFIM": 34, "HM": 34, "RFSM": 34, "RASM": 34, "RSM": 34,
                     "ISM": 34, "APSM": 34, "IBDM": 34,
                     "ABDM": 34, "TPIM": 34, "SPIM": 34, "TPSM": 34, "BLDC": 34, "ASSM": 34,
                     "PMDC": 34, "SRM": 34, "LSSM": 34, "UNIM": 34, "DCM": 34, "CPSM": 34, "NSSM": 34, "TR": 35,
                     "TransientTemperatureAndFlow": 36, "TransientTemperatureOnly": 37, "TransientFlowOnly": 38,
                     "Structural": 39}

    SetupTemplates = {0: HFSSDrivenAuto, 1: HFSSDrivenDefault, 2: HFSSEigen, 3: HFSSTransient, 4: HFSSSBR,
                      5: MaxwellTransient, 6: Magnetostatic, 7: EddyCurrent, 8: Electrostatic, 9: Electrostatic,
                      10: ElectricTransient, 11: SteadyTemperatureAndFlow, 12: SteadyTemperatureOnly,
                      13: SteadyFlowOnly, 14: Matrix, 15: NexximLNA, 16: NexximDC, 17: NexximTransient,
                      18: NexximQuickEye,
                      19: NexximVerifEye, 20: NexximAMI, 21: NexximOscillatorRSF, 22: NexximOscillator1T,
                      23: NexximOscillatorNT, 24: NexximHarmonicBalance1T, 25: NexximHarmonicBalanceNT,
                      26: NexximSystem, 27: NexximTVNoise, 28: HSPICE, 29: HFSS3DLayout, 30: Open, 31: Close,
                      32: MechTerm, 33: MechModal, 34: GRM, 35: TR, 36: TransientTemperatureAndFlow,
                      37: TransientTemperatureOnly, 38: TransientFlowOnly, 39: MechStructural}

    SetupNames = ["HFSSDrivenAuto", "HFSSDriven", "HFSSEigen", "HFSSTransient", "HFSSDriven",
                  "Transient", "Magnetostatic", "EddyCurrent", "Electrostatic", "ElectroDCConduction",
                  "ElectroDCConduction", "IcepakSteadyState", "IcepakSteadyState",
                  "IcepakSteadyState", "Matrix", "NexximLNA", "NexximDC", "NexximTransient", "NexximQuickEye",
                  "NexximVerifEye", "NexximAMI", "NexximOscillatorRSF", "NexximOscillator1T",
                  "NexximOscillatorNT", "NexximHarmonicBalance1T", "NexximHarmonicBalanceNT",
                  "NexximSystem", "NexximTVNoise", "HSPICE", "HFSS3DLayout", "2DMatrix", "2DMatrix", "MechThermal",
                  "MechModal", "GRM", "TR", "IcepakTransient", "IcepakTransient", "IcepakTransient", "MechStructural"]

    defaultAdaptive = {"DrivenModal": "LastAdaptive", "DrivenTerminal": "LastAdaptive", "Eigenmode": "LastAdaptive",
                       "Transient Network": "Transient", "SBR+": None, "Transient": "Transient",
                       "Magnetostatic": "LastAdaptive", "EddyCurrent": "LastAdaptive",
                       "Electrostatic": "LastAdaptive", "ElectroDCConduction": "LastAdaptive",
                       "ElectricTransient": "Transient", "SteadyTemperatureAndFlow": "SteadyState",
                       "SteadyTemperatureOnly": "SteadyState", "SteadyFlowOnly": "SteadyState",
                       "SteadyState": "SteadyState",
                       "Matrix": "LastAdaptive", "NexximLNA": None, "NexximDC": None, "NexximTransient": None,
                       "NexximQuickEye": None,
                       "NexximVerifEye": None, "NexximAMI": None, "NexximOscillatorRSF": None,
                       "NexximOscillator1T": None,
                       "NexximOscillatorNT": None, "NexximHarmonicBalance1T": None, "NexximHarmonicBalanceNT": None,
                       "NexximSystem": None, "NexximTVNoise": None, "HSPICE": None, "HFSS3DLayout": None,
                       "2DMatrix": "LastAdaptive", "MechThermal": "Solution", "MechModal": "Solution",
                       "GRM": "LastAdaptive", "TR": None, "TransientTemperatureAndFlow": "SteadyState",
                       "TransientTemperatureOnly": "Transient", "TransientFlowOnly": "Transient",
                       "Structural": "Solution"}
